""" JSON Module

"""
from abc import ABC, abstractmethod

import simplejson


class Serializable(ABC):
    """Python Interface to export Python objects to JSON format.

    The child class must implement 'dumps' property.
    """

    @abstractmethod
    def dumps(self) -> dict:
        """Property to generate the JSON formatted data of the current instance."""
        pass


def json_read(filepath) -> dict:
    """Read the JSON file with the given filepath.

    Parameters
    ----------
    filepath :

    Returns
    -------
    dict
        The JSON data of the given file as a dict or list.

    """
    with open(filepath, "rb") as json_file:
        json_data = simplejson.load(json_file)
        return json_data


def json_write(filepath, json_data, indent=4):
    """Write the given json_data to the JSON file with the given filepath.

    Parameters
    ----------
    filepath :
    json_data :
    indent :
    """
    with open(filepath, "w") as json_file:
        simplejson.dump(json_data, json_file, indent=indent)



import pandas as pd
import seaborn as sns
from scipy.interpolate import interp1d

sns.set_theme(style="whitegrid")
from math import pi

import matplotlib.pyplot as plt


# TODO: move
class PropertyInfo:
    def __init__(self, name):
        self.name = name


class PropertyValue(PropertyInfo):
    def __init__(self, name, best, worst):
        super().__init__(name)
        self.best = best
        self.worst = worst
        self.interp = interp1d(
            [worst, best], [0, 1], bounds_error=False, fill_value=(0, 1)
        )

    def interpolate(self, data):
        return self.interp(data)


# TODO: move
class PropertyName(PropertyInfo):
    def __init__(self):
        super().__init__("name")

    def compute(self, synapse, MTOC, perforin):
        return f"{synapse.cell_1.name}/{synapse.cell_2.name}"


class DataFrameBuilder:
    def __init__(self, index_property, group):
        self.index_property = index_property
        self.group = group
        self.data = {}
        self.data[self.index_property.name] = []
        self.data[self.group.name] = []
        self.properties = {}

    def add(self, cell_property):
        self.data[cell_property.name] = []
        self.properties[cell_property.name] = cell_property

    def compute(self, synapse, MTOC, perforin):
        self.data[self.index_property.name].append(
            self.index_property.compute(synapse, MTOC, perforin)
        )
        self.data[self.group.name].append(self.group.compute(synapse, MTOC, perforin))
        for name, p in self.properties.items():
            self.data[name].append(p.compute(synapse, MTOC, perforin))

    @property
    def dataframe(self):
        return pd.DataFrame(self.data)

    def interpolate(self):
        df = self.dataframe
        for k in self.properties.keys():
            df[k] = self.properties[k].interpolate(df[k])
        return df

    def show(self, filename=None):
        df = self.dataframe
        df_melt = pd.melt(
            df, value_vars=self.properties.keys(), id_vars=self.group.name
        )
        plt.figure(figsize=(3, 3))
        sns.violinplot(
            x="variable",
            y="value",
            hue=self.group.name,
            data=df_melt,
            palette="muted",
            split=True,
            inner="quartile",
        )
        if filename is not None:
            plt.savefig(filename)
        plt.show()

        """
        df = self.interpolate()
        df_melt = pd.melt(df, value_vars=self.properties.keys(), id_vars=self.group.name)
        plt.figure(figsize=(16, 9))
        sns.violinplot(x='variable', y='value', hue=self.group.name, data=df_melt, palette="muted", split=True,
                       inner="quartile")
        plt.show()
        """

    def show_radar(self):
        """
        categories = self.properties.keys()
        N = len(categories)
        # What will be the angle of each axis in the plot? (we divide the plot / number of variable)
        angles = [n / float(N) * 2 * pi for n in range(N)]
        angles += angles[:1]

        angle = np.deg2rad(67.5)
        figure = FigureSinglePolar(xlabel=categories, size=(4, 4))
        values = df.drop(columns=[self.index_property.name, self.group.name]).mean(axis=0).values.flatten().tolist()
        values += values[:1]

        figure.panel.plot(angles, values, linewidth=1, linestyle='solid', label="P")

        handles, labels = plt.gca().get_legend_handles_labels()
        handles, labels = zip(*[ (handles[i], labels[i]) for i in sorted(range(len(handles)), key=lambda k: labels[k])])
        figure.panel.legend(handles, labels, loc="lower left", bbox_to_anchor=(.5 + np.cos(angle)/2, .5 + np.sin(angle)/2))
        """
        """
        for i, j in df.head(20).iterrows():
            values=j.drop([self.index_property.name, self.group.name]).values.flatten().tolist()
            values += values[:1]

            if i == 0:
                figure.panel.plot(angles, values, linewidth=1, linestyle='dashed', label=f"maximum score", c='k')
            else:

            figure.panel.plot(angles, values, linewidth=1, linestyle='solid', label=f"pair {i}")
        """

        # number of variable
        categories = self.properties.keys()
        N = len(categories)

        # What will be the angle of each axis in the plot? (we divide the plot / number of variable)
        angles = [n / float(N) * 2 * pi for n in range(N)]
        angles += angles[:1]

        # Initialise the spider plot
        ax = plt.subplot(111, polar=True)

        # If you want the first axis to be on top:
        ax.set_theta_offset(pi / 2)
        ax.set_theta_direction(-1)

        # Draw one axe per variable + add labels
        plt.xticks(angles[:-1], categories)

        # Draw ylabels
        ax.set_rlabel_position(0)
        plt.yticks(
            [0.2, 0.4, 0.6, 0.8], ["0.2", "0.4", "0.6", "0.8"], color="grey", size=7
        )
        plt.ylim(0, 1)

        # ------- PART 2: Add plots

        # Plot each individual = each line of the data
        # I don't make a loop, because plotting more than 3 groups makes the chart unreadable

        # Ind1
        df = self.interpolate()
        values = (
            df[df[self.group.name] == "NP"]
            .drop(columns=[self.index_property.name, self.group.name])
            .mean(axis=0)
            .values.flatten()
            .tolist()
        )
        values += values[:1]
        ax.plot(angles, values, linewidth=1, linestyle="solid", label="NP")
        ax.fill(angles, values, "b", alpha=0.1)

        # Ind2
        values = (
            df[df[self.group.name] == "P"]
            .drop(columns=[self.index_property.name, self.group.name])
            .mean(axis=0)
            .values.flatten()
            .tolist()
        )
        values += values[:1]
        ax.plot(angles, values, linewidth=1, linestyle="solid", label="P")
        ax.fill(angles, values, "r", alpha=0.1)

        # Add legend
        plt.legend(loc="upper right", bbox_to_anchor=(0.1, 0.1))

        # Show the graph
        plt.show()



"""

"""

import cv2
import numpy as np
from skimage import color


def bgr2hsv(image):
    return cv2.cvtColor(image, cv2.COLOR_BGR2HSV)


def hsv2rgb(image):
    return cv2.cvtColor(image, cv2.COLOR_HSV2RGB)


def rgb2bgr(image):
    return cv2.cvtColor(image, cv2.COLOR_RGB2BGR)


def bgr2rgb(image):
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)


def gray2rgb(image):
    return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)


def rgb2gray(image):
    return cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)


def rgb2hed(image):
    return (color.rgb2hed(image) * 255).astype(np.uint8)


def bgr2hed(image):
    return rgb2hed(bgr2rgb(image))



"""

"""
import cv2
import numpy as np

from numena.enums import (
    IMAGE_UINT8_COLOR_1C,
    IMAGE_UINT8_COLOR_3C,
    IMAGE_UINT8_POSITIVE,
)
from numena.image.basics import image_new
from numena.image.contour import (
    contour_area,
    contours_convex_hull,
    contours_draw,
    contours_find,
    contours_merge,
)
from numena.io.image import imwrite


def draw_rectangle(image, x: int, y: int, width: int, height: int):
    """

    Parameters
    ----------
    image :
    x :
    y :
    width :
    height :
    """
    return cv2.rectangle(
        image, (x, y), (x + width, y + height), IMAGE_UINT8_COLOR_1C, -1
    )


def draw_overlay(image, mask, color=None, alpha=1.0, border_color="same", thickness=1):
    if color is None:
        color = IMAGE_UINT8_COLOR_3C
    out = image.copy()
    img_layer = image.copy()
    img_layer[np.where(mask)] = color
    overlayed = cv2.addWeighted(img_layer, alpha, out, 1 - alpha, 0, out)
    contours = contours_find(mask, exclude_holes=False)
    if border_color == "same":
        overlayed = contours_draw(overlayed, contours, thickness=thickness, color=color)
    elif border_color is not None:
        overlayed = contours_draw(
            overlayed, contours, thickness=thickness, color=border_color
        )
    return overlayed


def fill_ellipses(mask, ellipses, color=IMAGE_UINT8_POSITIVE):
    for ellipse in ellipses:
        cv2.ellipse(mask, ellipse, color, thickness=-1)
    return mask


def fill_ellipses_as_labels(mask, ellipses):
    for i, ellipse in enumerate(ellipses):
        cv2.ellipse(mask, ellipse, i + 1, thickness=-1)
    return mask


def fill_polygons(mask, polygons, color=IMAGE_UINT8_POSITIVE):
    return cv2.fillPoly(mask, pts=polygons, color=color)


def fill_polygons_as_labels(mask, polygons):
    for i, polygon in enumerate(polygons):
        cv2.fillPoly(mask, pts=np.int32([polygon]), color=i + 1)
    return mask


if __name__ == "__main__":
    image_test = image_new((512, 512))
    image_test = draw_rectangle(image_test, 256, 256, 64, 64)
    image_test = draw_rectangle(image_test, 16, 16, 64, 64)
    image_test = draw_rectangle(image_test, 92, 16, 16, 64)
    imwrite("rectangle_test.png", image_test)

    contours = contours_find(image_test)
    for cnt in contours:
        area = contour_area(cnt)
        print(area)

    merged_contour = contours_merge(contours)
    convex_hull = contours_convex_hull(merged_contour)
    area = contour_area(convex_hull)
    print(area)
    image_test = contours_draw(image_test, [convex_hull], color=[128], thickness=3)
    imwrite("rectangle_test_merged.png", image_test)



from __future__ import annotations

from dataclasses import asdict, astuple, dataclass
from math import atan2, cos, degrees, pi, radians, sin
from typing import ClassVar, Tuple

import cv2
import numpy as np

from numena.image.contour import contours_fill, contours_find
from numena.image.geometry import intersection_with_line
from numena.image.morphology import morph_dilate, morph_erode


@dataclass
class Vector2:
    x: float
    y: float
    ZERO: ClassVar[Vector2]
    UP: ClassVar[Vector2]
    DOWN: ClassVar[Vector2]
    LEFT: ClassVar[Vector2]
    RIGHT: ClassVar[Vector2]

    def __sub__(self, other: Vector2) -> Vector2:
        return Vector2(self.x - other.x, self.y - other.y)

    def __add__(self, other: Vector2) -> Vector2:
        return Vector2(self.x + other.x, self.y + other.y)

    def __mul__(self, scalar: float) -> Vector2:
        return Vector2(self.x * scalar, self.y * scalar)

    def dot(self, other: Vector2) -> float:
        return self.x * other.x + self.y * other.y

    def norm(self) -> float:
        return self.dot(self) ** 0.5

    def normalized(self) -> Vector2:
        norm = self.norm()
        return Vector2(self.x / norm, self.y / norm)

    def distance(self, other: Vector2) -> float:
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2) ** 0.5

    def perp(self) -> Vector2:
        return Vector2(1, -self.x / self.y)

    def as_tuple(self) -> Tuple[float, float]:
        return self.x, self.y

    def as_int_tuple(self) -> Tuple[int, int]:
        return int(round(self.x)), int(round(self.y))

    def angle_with_x_axis(self, other: Vector2) -> float:
        diff = other - self
        rad = atan2(diff.y, diff.x)
        if rad < 0:
            rad += 2 * pi
        return degrees(rad)

    def __str__(self) -> str:
        return str(astuple(self))

    def __repr__(self) -> str:
        return f"Vector2 {asdict(self)}"


Vector2.ZERO = Vector2(0.0, 0.0)
Vector2.UP = Vector2(0.0, 1.0)
Vector2.DOWN = Vector2(0.0, -1.0)
Vector2.LEFT = Vector2(-1.0, 0.0)
Vector2.RIGHT = Vector2(1.0, 0.0)


@dataclass
class Line2:
    pt1: Vector2
    pt2: Vector2
    ZERO: ClassVar[Line2]
    UP: ClassVar[Line2]
    DOWN: ClassVar[Line2]
    LEFT: ClassVar[Line2]
    RIGHT: ClassVar[Line2]

    @staticmethod
    def from_point(
        point: Vector2, angle: float, size: float, centered: bool = False
    ) -> Line2:
        angle_rad = radians(angle)
        angle_rad_inv = radians(angle - 180)
        if centered:
            size = size / 2
            point_1 = point + Vector2(cos(angle_rad_inv), sin(angle_rad_inv)) * size
        else:
            point_1 = point
        point_2 = point + Vector2(cos(angle_rad), sin(angle_rad)) * size
        return Line2(point_1, point_2)

    def as_int_tuple(self):
        return self.pt1.as_int_tuple(), self.pt2.as_int_tuple()


Line2.ZERO = Line2(Vector2.ZERO, Vector2.ZERO)
Line2.UP = Line2(Vector2.ZERO, Vector2.UP)
Line2.DOWN = Line2(Vector2.ZERO, Vector2.DOWN)
Line2.LEFT = Line2(Vector2.ZERO, Vector2.LEFT)
Line2.RIGHT = Line2(Vector2.ZERO, Vector2.RIGHT)


class MicroEntity:
    def __init__(self, name, custom_data=None):
        self.name = name
        self.custom_data = custom_data
        if self.custom_data is None:
            self.custom_data = {}

    def set_data(self, data_name, data):
        self.custom_data[data_name] = data

    def get_data(self, data_name):
        return self.custom_data[data_name]


class MicroEntity2D(Vector2, MicroEntity):
    def __init__(self, name, mask, x, y, custom_data=None):
        MicroEntity.__init__(self, name, custom_data=custom_data)
        Vector2.__init__(self, x, y)
        self.mask = mask

    def get_mean(self, channel):
        return np.mean(channel, where=self.mask > 0)

    def get_sum(self, channel, threshold=0):
        return np.sum(channel, where=self.mask > threshold)

    def get_max(self, channel):
        return np.max(channel, where=self.mask > 0, initial=0)

    @property
    def area(self):
        return cv2.countNonZero(self.mask)

    @property
    def boundary(self):
        if self.area == 0:
            return None
        return contours_find(self.mask)

    @property
    def perimeter(self):
        return cv2.arcLength(self.boundary[0], True)

    @property
    def roundness(self):
        return 4 * pi * (self.area / self.perimeter**2)

    @property
    def min_x(self):
        return np.min(self.boundary[0], axis=0)[0, 0]

    @property
    def max_x(self):
        return np.max(self.boundary[0], axis=0)[0, 0]


class Particle2D(MicroEntity2D):
    pass


class Cell2D(MicroEntity2D):
    def __init__(self, name, mask, x, y, custom_data=None):
        super().__init__(name, mask, x, y, custom_data)

    def dilated_mask(self, dilation_size=1):
        return morph_dilate(self.mask, "circle", half_kernel_size=dilation_size)

    def eroded_mask(self, dilation_size=1):
        return morph_erode(self.mask, "circle", half_kernel_size=dilation_size)

    @staticmethod
    def from_mask(cell_mask, cell_name, area_range=None, custom_data={}):
        mask = np.zeros(cell_mask.shape, dtype=np.uint8)
        cnts = contours_find(cell_mask)
        if len(cnts) == 1:
            cnt = cnts[0]
            if len(cnt) >= 4:
                M = cv2.moments(cnt)
                if M["m00"] == 0:
                    return None
                cx = int(M["m10"] / M["m00"])
                cy = int(M["m01"] / M["m00"])
                mask = contours_fill(mask, [cnt], color=255)

                if area_range is not None:
                    area = cv2.countNonZero(mask)
                    if area_range[0] <= area <= area_range[1]:
                        return Cell2D(cell_name, mask, cx, cy, custom_data=custom_data)
                    else:
                        return None
                else:
                    return Cell2D(cell_name, mask, cx, cy, custom_data=custom_data)
        return None


class Synapse(MicroEntity):
    def __init__(self, cell_1, cell_2, custom_data=None):
        super().__init__(f"{cell_1.name}-{cell_2.name}", custom_data)
        self.cell_1 = cell_1
        self.cell_2 = cell_2

    @property
    def angle(self):
        return self.cell_1.angle_with_x_axis(self.cell_2)

    @property
    def distance(self):
        return self.cell_1.distance(self.cell_2)

    def front_cell_1(self):
        line = [self.cell_1.as_int_tuple(), self.cell_2.as_int_tuple()]
        return intersection_with_line(self.cell_1.mask, line)

    def crop(self, padding=5):
        synapse_mask = self.cell_1.mask | self.cell_2.mask
        x, y, w, h = cv2.boundingRect(synapse_mask)
        x = max(x - padding, 0)
        y = max(y - padding, 0)
        w = w + padding * 2
        h = h + padding * 2
        return x, y, w, h



"""Image Module

"""

import cv2
import numpy as np


def image_new(shape, dtype=np.uint8):
    return np.zeros(shape=shape, dtype=dtype)


def image_like(image):
    return np.zeros_like(image)


def image_split(image):
    return list(cv2.split(image))


def image_ew_mean(image_1, image_2):
    return cv2.addWeighted(image_1, 0.5, image_2, 0.5, 0)


def image_ew_max(image_1, image_2):
    return cv2.max(image_1, image_2)


def image_ew_min(image_1, image_2):
    return cv2.min(image_1, image_2)


def image_normalize(image):
    return cv2.normalize(
        image, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
    )



import cv2
import czifile
import tifffile

from numena.image.color import bgr2rgb


def imread_color(filename, rgb=False):
    if rgb:
        return bgr2rgb(cv2.imread(filename, cv2.IMREAD_COLOR))
    return cv2.imread(filename, cv2.IMREAD_COLOR)


def imread_grayscale(filename):
    return cv2.imread(filename, cv2.IMREAD_GRAYSCALE)


def imread_mirax(filename):
    pass


def imread_lsm(filename):
    return tifffile.imread(filename)


def imread_tiff(filename):
    return tifffile.imread(filename)


def imread_czi(filename):
    return czifile.imread(filename)


def imwrite(filename, image):
    cv2.imwrite(filename, image)


def imwrite_tiff(filename, image, imagej=True, luts=None):
    if luts:
        tifffile.imwrite(
            filename,
            image,
            imagej=imagej,
            metadata={"mode": "composite"},
            ijmetadata=luts,
        )



import itertools

import numpy as np

from numena.io.json import json_read


class ProfilingInfo:
    def __init__(self, filename):
        json_data = json_read(filename)
        self.method = json_data["method"]
        self.channels = {}
        for c in json_data["channels"].keys():
            self.channels[int(c)] = json_data["channels"][c]
        self.separator = json_data["separator"]
        self.conditions = json_data["conditions"]


class CellStainingProfile:
    def __init__(self, infos):
        self.infos = infos
        self.lst = list(itertools.product([0, 1], repeat=len(infos.channels)))
        self.lst = sorted(self.lst, key=lambda c: sum(c), reverse=True)

    def get_profile(self, cell, stainings):
        inside_cell = cell.mask > 0
        labels = []
        profile = []
        channels_true = {}
        channels_false = {}
        for i, channel in self.infos.channels.items():
            channel_data = stainings[:, i]
            name = channel["name"]
            threshold = channel["threshold"]
            channel_true = channel_data >= threshold[0]
            channel_false = ~channel_true
            channels_true[name] = channel_true & inside_cell
            channels_false[name] = channel_false & inside_cell

        for combination in self.lst:
            combination_names = []
            combination_values = []
            for i, channel in self.infos.channels.items():
                name = channel["name"]
                if combination[i] == 1:
                    combination_names.append(name)
                    combination_values.append(channels_true[name])
                else:
                    combination_values.append(channels_false[name])
            s = np.count_nonzero(np.logical_and.reduce(combination_values))
            label_name = f"{self.infos.separator}".join(combination_names)
            labels.append(label_name)
            profile.append(s)

        for i, channel in self.infos.channels.items():
            name = channel["name"]
            labels.append(f"{name}+")
            profile.append(np.count_nonzero(channels_true[name]))
            labels.append(f"{name}-SUM")
            profile.append(np.sum(stainings[:, i], where=channels_true[name]))
            labels.append(f"{name}-MEAN")
            profile.append(np.mean(stainings[:, i], where=channels_true[name]))
        return profile, labels



"""Morphology Module

"""

import cv2
import numpy as np
from scipy import ndimage
from skimage.feature import peak_local_max
from skimage.segmentation import watershed

from numena.image.contour import contours_fill, contours_find


class WatershedTransform:
    pass


class WatershedSkimage(WatershedTransform):
    def __init__(self, use_dt=False, markers_distance=21, markers_area=None):
        self.use_dt = use_dt
        self.markers_distance = markers_distance
        self.markers_area = markers_area

    def _extract_markers(self, signal):
        peak_idx = peak_local_max(
            signal,
            min_distance=self.markers_distance,
            exclude_border=0,
            # footprint=np.ones((21, 21)),
            num_peaks=255,
        )
        peak_mask = np.zeros_like(signal, dtype=np.uint8)
        peak_mask[tuple(peak_idx.T)] = 1
        return peak_mask

    def apply(self, signal, markers=None, mask=None):
        if self.use_dt:
            signal = ndimage.distance_transform_edt(signal)
        if markers is None:
            markers = self._extract_markers(signal)

        # markers[mask == 0] = 0
        if self.markers_area:
            n, marker_labels, stats, _ = cv2.connectedComponentsWithStats(
                markers, connectivity=8
            )
            for i in range(1, n):
                if (
                    self.markers_area[0]
                    < stats[i, cv2.CC_STAT_AREA]
                    < self.markers_area[1]
                ):
                    pass
                else:
                    marker_labels[marker_labels == i] = 0
        else:
            marker_labels = cv2.connectedComponents(
                markers, connectivity=8, ltype=cv2.CV_16U
            )[1]

        # limit the number of markers to 255
        marker_labels[marker_labels > 255] = 0
        signal_inv = 255 - signal
        labels = watershed(
            signal_inv, markers=marker_labels, mask=mask, watershed_line=True
        )
        return signal, marker_labels, labels


def get_kernel(struct_shape="circle", half_kernel_size=1):
    kernel_size = half_kernel_size * 2 + 1
    if struct_shape == "circle":
        return cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (kernel_size, kernel_size))
    elif struct_shape == "square":
        return cv2.getStructuringElement(cv2.MORPH_RECT, (kernel_size, kernel_size))
    elif struct_shape == "cross":
        return cv2.getStructuringElement(cv2.MORPH_CROSS, (kernel_size, kernel_size))


def morph_erode(image, struct_shape="circle", half_kernel_size=1):
    kernel = get_kernel(struct_shape=struct_shape, half_kernel_size=half_kernel_size)
    return cv2.erode(image, kernel)


def morph_dilate(image, struct_shape="circle", half_kernel_size=1):
    kernel = get_kernel(struct_shape=struct_shape, half_kernel_size=half_kernel_size)
    return cv2.dilate(image, kernel)


def morph_fill(image):
    cnts = contours_find(image, exclude_holes=True)
    return contours_fill(image, cnts)



from dataclasses import dataclass
from math import pi

import matplotlib.pyplot as plt


@dataclass
class Figure:
    title: str = None
    size: tuple = (12, 12)
    dpi: int = 300
    background_color: tuple = (1, 1, 1)
    hspace: float = 0.25

    def create_panels(self, rows=2, cols=2, show_axis=True):
        self.__fig, self.panels = plt.subplots(
            nrows=rows, ncols=cols, figsize=self.size, dpi=self.dpi
        )
        if self.title:
            self.__fig.suptitle(self.title)
        self.__fig.set_facecolor(self.background_color)
        self.__fig.subplots_adjust(hspace=self.hspace)
        self.panels = self.panels.ravel()
        if not show_axis:
            [p.axis("off") for p in self.panels]

    def set_panel(self, n, title, image, cmap=None, hide_axis=True):
        panel = self.get_panel(n)
        panel.set_title(title)
        if hide_axis:
            panel.axis("off")
        panel.imshow(image, cmap=cmap)

    def get_panel(self, n):
        return self.panels[n]

    def set_projection(self, n, projection):
        self.panels[n].remove()
        self.panels[n] = self.__fig.add_subplot(3, 2, n + 1, projection=projection)

    def save(self, filename):
        self.__fig.savefig(filename)

    def close(self):
        plt.close(self.__fig)


@dataclass
class FigureSinglePanel:
    title: str = None
    xlabel: str = None
    ylabel: str = None
    size: tuple = (12, 12)
    dpi: int = 300
    background_color: tuple = (1, 1, 1)
    left: float = 0.125
    right: float = 0.875

    def __post_init__(self):
        self.__fig, self.panel = plt.subplots(
            nrows=1, ncols=1, figsize=self.size, dpi=self.dpi
        )
        self.__fig.subplots_adjust(left=self.left, right=self.right)
        if self.title:
            self.__fig.suptitle(self.title)
        if self.xlabel:
            self.panel.set_xlabel(self.xlabel)
        if self.ylabel:
            self.panel.set_ylabel(self.ylabel)
        self.__fig.set_facecolor(self.background_color)

    def get_panel(self):
        return self.panel

    def set_projection(self, n, projection):
        self.panels[n].remove()
        self.panels[n] = self.__fig.add_subplot(3, 2, n + 1, projection=projection)

    def save(self, filename):
        self.__fig.savefig(filename)


@dataclass
class FigureSingleImage(FigureSinglePanel):
    def __post_init__(self):
        super().__post_init__()
        self.panel.axis("off")

    def get_panel(self):
        return self.panel

    def set_projection(self, n, projection):
        self.panels[n].remove()
        self.panels[n] = self.__fig.add_subplot(3, 2, n + 1, projection=projection)

    def save(self, filename):
        self.__fig.savefig(filename)


@dataclass
class FigureSinglePolar(FigureSinglePanel):
    def __post_init__(self):
        self.__fig, self.panel = plt.subplots(
            nrows=1,
            ncols=1,
            figsize=self.size,
            dpi=self.dpi,
            subplot_kw={"projection": "polar"},
        )
        N = len(self.xlabel)
        # If you want the first axis to be on top:
        self.panel.set_theta_offset(pi / 2)
        self.panel.set_theta_direction(-1)

        self.angles = [n / float(N) * 2 * pi for n in range(N)]
        self.angles += self.angles[:1]

        # Draw ylabels
        self.panel.set_rlabel_position(0)
        self.panel.set_rticks([0.25, 0.5, 0.75])  # Less radial ticks
        self.panel.set_rmax(1)
        if self.title:
            self.__fig.suptitle(self.title)
        if self.xlabel:
            self.panel.set_xticks(self.angles[:-1], minor=False)
            self.panel.set_xticklabels(self.xlabel, fontdict=None, minor=False)
        self.panel.plot(self.angles, [0] * (N + 1), linewidth=0)
        self.__fig.set_facecolor(self.background_color)

    def get_panel(self):
        return self.panel

    def save(self, filename):
        self.__fig.savefig(filename)



"""Contours Module

"""

import cv2
import numpy as np

from numena.enums import IMAGE_UINT8_COLOR_1C, IMAGE_UINT8_COLOR_3C


def contours_convex_hull(contour):
    return cv2.convexHull(contour)


def contours_merge(contours):
    return np.vstack(contours)


def contour_area(contour):
    return cv2.contourArea(contour)


def contours_find(image, exclude_holes=True):
    mode = cv2.RETR_LIST
    method = cv2.CHAIN_APPROX_SIMPLE
    if exclude_holes:
        mode = cv2.RETR_EXTERNAL
    return cv2.findContours(image.copy(), mode, method)[0]


def _draw_contours(
    image, contours, color=None, selected: int = None, fill=True, thickness=1
):
    assert (
        len(image.shape) == 3 or len(image.shape) == 2
    ), "given image wrong format, shape must be (h, w, c) or (h, w)"
    if color is None:
        if len(image.shape) == 3 and image.shape[-1] == 3:
            color = IMAGE_UINT8_COLOR_3C
        elif len(image.shape) == 2:
            color = IMAGE_UINT8_COLOR_1C
        else:
            raise ValueError("Image wrong format, must have 1 or 3 channels")
    if selected is None:
        selected = -1  # selects all the contours (-1)
    if fill:
        thickness = -1  # fills the contours (-1)
    return cv2.drawContours(image.copy(), contours, selected, color, thickness)


def contours_fill(image, contours, color=None, selected=None):
    return _draw_contours(image, contours, color, selected, fill=True)


def contours_draw(image, contours, color=None, selected=None, thickness=1):
    return _draw_contours(
        image, contours, color, selected, fill=False, thickness=thickness
    )



"""

"""

from roifile import ImagejRoi


def read_ellipses_from_csv(dataframe, scale=1.0, ellipse_scale=1.0):
    ellipses = []
    for _, ellipse in dataframe.iterrows():
        ellipse_info = read_ellipse_from_row(ellipse, scale, ellipse_scale)
        ellipses.append(ellipse_info)
    return ellipses


def read_ellipse_from_row(row, scale=1.0, ellipse_scale=1.0):
    angle = 180 - row.Angle
    position = (row.X * scale, row.Y * scale)
    size = (row.Major * scale * ellipse_scale, row.Minor * scale * ellipse_scale)
    ellipse_info = (position, size, angle)
    return ellipse_info


def read_polygons_from_roi(filename, scale=1.0):
    rois = ImagejRoi.fromfile(filename)
    if type(rois) == ImagejRoi:
        return [rois.coordinates()]
    polygons = [roi.coordinates() for roi in rois]
    return polygons



"""Threshold Module

"""

import cv2

from numena.enums import IMAGE_UINT8_POSITIVE


def threshold_binary(image, threshold, value=IMAGE_UINT8_POSITIVE):
    return cv2.threshold(image, threshold, value, cv2.THRESH_BINARY)[1]


def threshold_tozero(image, threshold):
    return cv2.threshold(image, threshold, IMAGE_UINT8_POSITIVE, cv2.THRESH_TOZERO)[1]



"""Time Module

"""

from datetime import datetime
from uuid import uuid4


def timestamp(ms=True):
    """

    Parameters
    ----------
    ms :

    Returns
    -------

    """
    dt = datetime.now()
    if ms:
        return dt.microsecond
    return dt


def uuid():
    """

    Returns
    -------

    """
    return str(uuid4())


def eventid():
    """

    Returns
    -------

    """
    return f"{timestamp()}-{uuid()}".replace(" ", "-")



"""Drive Module

"""

import zipfile
from abc import ABC
from dataclasses import InitVar, dataclass, field
from pathlib import Path

import pandas as pd

from numena.io.image import imread_color, imread_czi, imread_tiff, imwrite, imwrite_tiff

EXTENSION_IMAGE_PNG = ".png"
EXTENSION_IMAGE_JPG = ".jpg"
CSV = ".csv"
EXTENSION_IMAGE_TIF = ".tif"
ZIP = ".zip"
EXTENSION_IMAGE_LSM = ".lsm"
EXTENSION_IMAGE_CZI = ".czi"


@dataclass
class DriveEntity(ABC):
    path: InitVar[str]
    _path: Path = field(init=False)

    def __post_init__(self, path):
        self._path = Path(path)

    def __getattr__(self, attr):
        return getattr(self._path, attr)

    def __truediv__(self, key):
        return self._path / key


@dataclass
class Directory(DriveEntity):
    def __post_init__(self, path):
        super().__post_init__(path)
        if not self.exists():
            _err = f"The directory {self._path} does not exist!"
            raise ValueError(_err)
        if not self.is_dir():
            _err = f"The path {self._path} is not pointing to a directory!"
            raise ValueError(_err)

    def write(self, filename, filedata):
        filepath = self / filename
        extension = filepath.suffix
        filepath = str(filepath)
        if extension == EXTENSION_IMAGE_PNG:
            imwrite(filepath, filedata)
        if extension == EXTENSION_IMAGE_TIF or extension == EXTENSION_IMAGE_LSM:
            imwrite_tiff(filepath, filedata)

    def read(self, filename):
        filepath = self / filename
        if not filepath.exists():
            _err = f"The file {filepath} does not exist!"
            raise ValueError(_err)
        if not filepath.is_file():
            _err = f"The path {filepath} is not pointing to a file!"
            raise ValueError(_err)
        extension = filepath.suffix
        filepath = str(filepath)
        if extension == EXTENSION_IMAGE_PNG or extension == EXTENSION_IMAGE_JPG:
            return imread_color(filepath)
        if extension == CSV:
            return pd.read_csv(filepath)
        if extension == EXTENSION_IMAGE_TIF or extension == EXTENSION_IMAGE_LSM:
            return imread_tiff(filepath)
        if extension == EXTENSION_IMAGE_CZI:
            return imread_czi(filepath)

    def unzip(self, filename):
        filepath = self / filename
        extension = filepath.suffix
        filepath = str(filepath)
        unzip_folder = filename.replace(ZIP, "")
        new_location = self / unzip_folder
        if new_location.exists():
            return Directory(new_location)

        if extension == ZIP:
            with zipfile.ZipFile(filepath, "r") as zip_ref:
                zip_ref.extractall(str(new_location))
            return Directory(new_location)

    def next(self, next_location):
        filepath = self / next_location
        filepath.mkdir(parents=True, exist_ok=True)
        return Directory(filepath)

    def ls(self, regex="*", ordered=False):
        if ordered:
            return sorted(self.glob(regex))
        return self.glob(regex)

    def save_as_csv(self, data, filename):
        filepath = self / filename
        if isinstance(data, dict):
            df = pd.DataFrame(data=data)
        elif isinstance(data, pd.DataFrame):
            df = data
        df.to_csv(str(filepath))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['numena', 'numena.dataset', 'numena.features', 'numena.image', 'numena.io']

package_data = \
{'': ['*']}

install_requires = \
['czifile',
 'matplotlib',
 'numpy',
 'opencv-python',
 'pandas',
 'roifile',
 'scikit-image',
 'simplejson',
 'tifffile']

setup_kwargs = {
    'name': 'numena',
    'version': '0.1.3',
    'description': 'Image Processing and Analysis toolbox package',
    'long_description': '# Numena\n',
    'author': 'Kevin Cortacero',
    'author_email': 'cortacero@inserm.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import cv2
import numpy as np

from numena.enums import IMAGE_UINT8_POSITIVE
from numena.image.basics import image_new
from numena.image.contour import contours_find


def intersection_with_line(mask, line):
    line_mask = image_new(mask.shape)
    line_mask = cv2.line(line_mask, line[0], line[1], IMAGE_UINT8_POSITIVE, 2)
    contours = contours_find(mask)
    mask_cnt = image_new(mask.shape)
    cv2.drawContours(mask_cnt, contours, -1, IMAGE_UINT8_POSITIVE, 2)
    intersection = cv2.bitwise_and(line_mask, mask_cnt)
    centroid = np.mean(np.argwhere(intersection), axis=0)
    return centroid
